#include <sys/libkern.h>
#include <sys/types.h>
#include <sys/systm.h>
#include <sys/param.h>
#include <sys/kernel.h>
#include <sys/malloc.h>
#include <sys/time.h>
#include <sys/bus.h>
#include <sys/proc.h>
#include <pci/pcireg.h>
#include <pci/pcivar.h>
#include <machine/ansi.h>
#include <machine/clock.h>
#include <vm/vm.h>
#include <vm/vm_object.h>
#include <vm/pmap.h>
#include <machine/pmap.h>
#include <machine/stdarg.h>
#define RM_HEADERS_MINIMAL
#include "nvrm.h"
#include "os-interface.h"
#include "nv-freebsd.h"

char *
osStringCopy (char *dst, const char *src)
{
	dvprintf ("osStringCopy\n");
	return (strcpy (dst, src));
}

RM_STATUS
osStringNCopyFromUser (char *dst, const char *src, int n)
{
	int done;
	dvprintf ("osStringNCopyFromUser\n");
	return (copyinstr (src, dst, n, &done) ? RM_ERR_BAD_ADDRESS : RM_OK);
}

S032
osStringCompare (const char *s1, const char *s2)
{
	dvprintf ("osStringCompare\n");
	return (strcmp (s1, s2));
}

U032
osStringLength (const char *str)
{
	dvprintf ("osStringLength\n");
	return (strlen (str));
}

RM_STATUS
osSetMemRange (U032 start, U032 size, U032 mode)
{
	return RM_OK;
}

RM_STATUS
osUnsetMemRange (U032 start, U032 size)
{
	return RM_ERROR;
}

void *
osMemCopy (void *dst, const void *src, U032 length)
{
	dvprintf ("osMemCopy\n");
	return (memcpy (dst, src, length));
}

BOOL osIsAdministrator(PHWINFO pDev)
{
    return suser(curproc);
}

RM_STATUS
osMemCopyFromUser (void *dst, const void *src, U032 length)
{
	dvprintf ("osMemCopyFromUser %ld\n", length);
	return copyin (src, dst, length) ? RM_ERR_BAD_ADDRESS : RM_OK;
}

RM_STATUS
osMemCopyToUser (void *dst, const void *src, U032 length)
{
	dvprintf ("osMemCopyToUser %ld %p\n", length, dst);
	return copyout (src, dst, length) ? RM_ERR_BAD_ADDRESS : RM_OK;
}

void *
osMemSet (void *dst, int c, U032 length)
{
	dvprintf ("osMemSet\n");
	return (memset (dst, c, length));
}

int
osMemCmp (const void *buf0, const void *buf1, U032 length)
{
	dvprintf ("osMemCmp\n");
	return (memcmp (buf0, buf1, length));
}

/* Operating System Memory functions
 */

MALLOC_DEFINE(M_NV, "nvidia", "nvidia memory store");

RM_STATUS
osAllocMem (void **pAddress, U032 size)
{
	dvprintf ("osAllocMem: %ld\n", size);
	*pAddress = malloc (size, M_NV, M_NOWAIT);
	dvprintf ("Address: %p\n", *pAddress);
	if (*pAddress == NULL)
		return RM_ERR_NO_FREE_MEM;
	return (RM_OK);
}


RM_STATUS
osFreeMem (void *pAddress)
{
/*	printf ("osFreeMem\n");*/
	free (pAddress, M_NV);
	return (RM_OK);
}

#ifdef DEBUG
#endif /* DEBUG */

RM_STATUS
osGetCurrentTime (U032 *seconds, U032 *useconds)
{
	struct timeval tv;

	microtime (&tv);
	*seconds = tv.tv_sec;
	*useconds = tv.tv_usec;

	return (RM_OK);
}

RM_STATUS osGetCurrentProcess (U032 *pPid);

RM_STATUS
osGetCurrentProcess (U032 *pPid)
{
	dprintf ("osGetCurrentProcess");
	*pPid = curproc->p_pid;

	return RM_OK;
}

RM_STATUS
osDelay (U032 MilliSeconds)
{
	unsigned int delay;
	struct timeval tv;

	dprintf ("osDelay");
	tv.tv_sec = 0;
	tv.tv_usec = MilliSeconds * 1000;

	delay = tvtohz (&tv);
	tsleep ("nv", PPAUSE, "nvidia", delay);

	return (RM_OK);
}

RM_STATUS osDelayUs (U032 MicroSeconds);

RM_STATUS
osDelayUs (U032 MicroSeconds)
{
	unsigned int	delay;
	struct timeval tv;

	dprintf ("osDelayUs\n");
	tv.tv_sec = 0;
	tv.tv_usec = MicroSeconds;

	delay = tvtohz (&tv);

	tsleep ("nv", PPAUSE, "nvidia", delay);

	return (RM_OK);
}

U032
osGetCpuFrequency (void)
{
	dprintf ("osGetCpuFrequency\n");
	return (tsc_freq+4999)*0.000001;
	return (0);
}

RM_STATUS
osKillProcess (U032 pid, U032 sig)
{
	dprintf ("osKillProcess\n");
	return (RM_ERROR);
}

const char *nv_rm_status_text (int local_errno);

static void
out_string (const char *str)
{
	dprintf ("%s", str);
}

const char *
nv_rm_status_text (int local_errno)
{
	static char buff[32];
	static struct {
		unsigned int key;
		char *text;
	} *ep, errs[] = {
	{ RM_OK, 			"RM_OK"},
	{ RM_ERROR,			"RM_ERROR"},
	{ NV_TYPE_CONFLICT,		"NV_TYPE_CONFLICT"},
	{ NV_OUT_OF_RANGE,		"NV_OUT_OF_RANGE"},
	{ NV_NO_CONNECTION,		"NV_NO_CONNECTION"},
	{ NV_NO_SUCH_OBJECT,		"NV_NO_SUCH_OBJECT" },
	{ NV_NAME_IN_USE,		"NV_NAME_IN_USE" },
	{ NV_OUT_OF_RESOURCES,		"NV_OUT_OF_RESOURCES" },
	{ NV_TRANSLATION_VIOLATION,	"NV_TRANSLATION_VIOLATION" },
	{ NV_PROTECTION_VIOLATION,	"NV_PROTECTION_VIOLATION" },
	{ NV_BUFFER_BUSY,		"NV_BUFFER_BUSY" },
	{ NV_ILLEGAL_ACCESS,		"NV_ILLEGAL_ACCESS" },
	{ NV_BAD_COLORMAP_FORMAT,	"NV_BAD_COLORMAP_FORMAT" },

	{ NV_BAD_COLOR_FORMAT,		"NV_BAD_COLOR_FORMAT" },
	{ NV_BAD_MONOCHROME_FORMAT,	"NV_BAD_MONOCHROME_FORMAT" },
	{ NV_BAD_PATTERN_SHAPE,		"NV_BAD_PATTERN_SHAPE" },
	{ NV_BAD_SUBDIVIDE,		"NV_BAD_SUBDIVIDE" },
	{ NV_NO_CURRENT_POINT,		"NV_NO_CURRENT_POINT" },
	{ NV_BAD_AUDIO_FORMAT,		"NV_BAD_AUDIO_FORMAT" },
	{ NV_NO_DMA_TRANSLATION,	"NV_NO_DMA_TRANSLATION" },
	{ NV_INCOMPLETE_METHOD,		"NV_INCOMPLETE_METHOD" },
	{ NV_RESERVED_ADDRESS,		"NV_RESERVED_ADDRESS" },
	{ NV_UNIMPLEMENTED_PATCH,	"NV_UNIMPLEMENTED_PATCH" },
	{ NV_OS_NAME_ERROR,		"NV_OS_NAME_ERROR" },
	{ NV_INCOMPLETE_PATCH,		"NV_INCOMPLETE_PATCH" },
	{ NV_BUFFERGAP_ERROR,		"NV_BUFFERGAP_ERROR" },
	{ NV_FIFO_OVERFLOW,		"NV_FIFO_OVERFLOW" },
	{ RM_WARN_NULL_OBJECT,		"RM_WARN_NULL_OBJECT" },
	{ RM_WARN_AUDIO_DISABLED,	"RM_WARN_AUDIO_DISABLED" },
	{ RM_WARN_GRAPHICS_DISABLED,	"RM_WARN_GRAPHICS_DISABLED" },
	{ RM_ERR_NO_FREE_MEM,		"RM_ERR_NO_FREE_MEM" },
	{ RM_ERR_MEM_NOT_FREED,		"RM_ERR_MEM_NOT_FREED" },
	{ RM_ERR_PAGE_TABLE_NOT_AVAIL,	"RM_ERR_PAGE_TABLE_NOT_AVAIL" },
	{ RM_ERR_NO_FREE_FIFOS,		"RM_ERR_NO_FREE_FIFOS" },
	{ RM_ERR_CANT_CREATE_CLASS_OBJS,"RM_ERR_CANT_CREATE_CLASS_OBJS" },
	{ RM_ERR_BAD_OBJECT,		"RM_ERR_BAD_OBJECT" },
	{ RM_ERR_INSERT_DUPLICATE_NAME,	"RM_ERR_INSERT_DUPLICATE_NAME" },
	{ RM_ERR_OBJECT_NOT_FOUND,	"RM_ERR_OBJECT_NOT_FOUND" },
	{ RM_ERR_CREATE_BAD_CLASS,	"RM_ERR_CREATE_BAD_CLASS" },
	{ RM_ERR_DELETE_BAD_CLASS,	"RM_ERR_DELETE_BAD_CLASS" },
	{ RM_ERR_FIFO_RUNOUT_OVERFLOW,	"RM_ERR_FIFO_RUNOUT_OVERFLOW" },
	{ RM_ERR_FIFO_BAD_ACCESS,	"RM_ERR_FIFO_BAD_ACCESS" },
	{ RM_ERR_FIFO_OVERFLOW,		"RM_ERR_FIFO_OVERFLOW" },
	{ RM_ERR_METHOD_ORDER,		"RM_ERR_METHOD_ORDER" },
	{ RM_ERR_METHOD_COUNT,		"RM_ERR_METHOD_COUNT" },
	{ RM_ERR_ILLEGAL_OBJECT,	"RM_ERR_ILLEGAL_OBJECT" },
	{ RM_ERR_DMA_IN_USE,		"RM_ERR_DMA_IN_USE" },
	{ RM_ERR_BAD_DMA_SPECIFIER,	"RM_ERR_BAD_DMA_SPECIFIER" },
	{ RM_ERR_INVALID_XLATE,		"RM_ERR_INVALID_XLATE" },
	{ RM_ERR_INVALID_START_LENGTH,	"RM_ERR_INVALID_START_LENGTH" },
	{ RM_ERR_DMA_MEM_NOT_LOCKED,	"RM_ERR_DMA_MEM_NOT_LOCKED" },
	{ RM_ERR_DMA_MEM_NOT_UNLOCKED,	"RM_ERR_DMA_MEM_NOT_UNLOCKED" },
	{ RM_ERR_NOTIFY_IN_USE,		"RM_ERR_NOTIFY_IN_USE" },
	{ RM_ERR_ILLEGAL_ADDRESS,	"RM_ERR_ILLEGAL_ADDRESS" },
	{ RM_ERR_BAD_ADDRESS,		"RM_ERR_BAD_ADDRESS" },
	{ RM_ERR_INVALID_COLOR_FORMAT,	"RM_ERR_INVALID_COLOR_FORMAT" },
	{ RM_ERR_INVALID_MONO_FORMAT,	"RM_ERR_INVALID_MONO_FORMAT" },
	{ RM_ERR_INVALID_AUDIO_FORMAT,	"RM_ERR_INVALID_AUDIO_FORMAT" },
	{ RM_ERR_INVALID_GAMEPORT_FORMAT,"RM_ERR_INVALID_GAMEPORT_FORMAT" },
	{ RM_ERR_OBJECT_TYPE_MISMATCH,	"RM_ERR_OBJECT_TYPE_MISMATCH" },
	{ RM_ERR_NO_FREE_AUDIO_INSTANCES,"RM_ERR_NO_FREE_AUDIO_INSTANCES" },
	{ RM_ERR_CODEC_INPUT_IN_USE,	"RM_ERR_CODEC_INPUT_IN_USE" },
	{ RM_ERR_AUDIO_DISABLED,	"RM_ERR_AUDIO_DISABLED" },
	{ RM_ERR_INCOMPLETE_PATCH,	"RM_ERR_INCOMPLETE_PATCH" },
	{ RM_ERR_INVALID_PATCH,		"RM_ERR_INVALID_PATCH" },
	{ RM_ERR_PATCH_TOO_COMPLEX,	"RM_ERR_PATCH_TOO_COMPLEX" },
	{ RM_ERR_MAX_PATCH_FANOUT,	"RM_ERR_MAX_PATCH_FANOUT" },
	{ RM_ERR_DEVICE_DISCONNECTED,	"RM_ERR_DEVICE_DISCONNECTED" },
#if 0
    /* whence NV_BUFFER_GAP? */
    { RM_ERR_BUFFER_GAP,                "RM_ERR_BUFFER_GAP" },
#endif
	{ RM_ERR_INVALID_OS_NAME,	"RM_ERR_INVALID_OS_NAME" },
	{ RM_ERR_ILLEGAL_ACTION,	"RM_ERR_ILLEGAL_ACTION" },
	{ RM_ERR_ILLEGAL_DIRECTION,	"RM_ERR_ILLEGAL_DIRECTION" },
	{ RM_ERR_OUT_OF_TIMER_CALLBACKS,"RM_ERR_OUT_OF_TIMER_CALLBACKS" },
	{ RM_ERR_OPERATING_SYSTEM,	"RM_ERR_OPERATING_SYSTEM" },
	{ RM_ERR_BAD_OBJECT_PARENT,	"RM_ERR_BAD_OBJECT_PARENT" },
	{ RM_ERR_BAD_OBJECT_HANDLE,	"RM_ERR_BAD_OBJECT_HANDLE" },
	{ RM_ERR_OBJECT_IN_USE,		"RM_ERR_OBJECT_IN_USE" },
	{ RM_ERR_OBJECT_HAS_CHILDERN,	"RM_ERR_OBJECT_HAS_CHILDERN" },
	{ RM_ERR_BAD_CLASS,		"RM_ERR_BAD_CLASS" },
	{ RM_ERR_INSUFFICIENT_RESOURCES,"RM_ERR_INSUFFICIENT_RESOURCES" },
	{ RM_ERR_BAD_FLAGS,		"RM_ERR_BAD_FLAGS" },
	{ RM_ERR_BAD_BASE,		"RM_ERR_BAD_BASE" },
	{ RM_ERR_BAD_LIMIT,		"RM_ERR_BAD_LIMIT" },
	{ RM_ERR_PROTECTION_FAULT,	"RM_ERR_PROTECTION_FAULT" },
	{ RM_ERR_MULTIPLE_MEMORY_TYPES,	"RM_ERR_MULTIPLE_MEMORY_TYPES" },
	{ RM_ERR_BAD_OBJECT_ERROR,	"RM_ERR_BAD_OBJECT_ERROR" },
	{ RM_ERR_BAD_OBJECT_BUFFER,	"RM_ERR_BAD_OBJECT_BUFFER" },
	{ RM_ERR_BAD_OFFSET,		"RM_ERR_BAD_OFFSET" },
	{ 0, 0 },
	};

	dprintf ("nv_rm_status_text\n");

	for (ep = errs; ep->text; ep++)
		if ((ep->key & ~NV_ERROR_MASK) == local_errno)
			return ep->text;

	sprintf (buff, "unknown error 0x%x", local_errno);
	return (buff);
}

int cur_debuglevel = DBG_LEVEL_INFO;

VOID
osDbgInit (void)
{
	U032 new_debuglevel;

	dprintf ("osDbgInit\n");

	if (RM_OK == osReadRegistryDword ((PHWINFO) 0,
					"",
					"resman_debuglevel",
					&new_debuglevel))
	{
		if (new_debuglevel != ~0)
			cur_debuglevel= new_debuglevel;
	}
}

#define MAX_ERROR_STRING 256

static int
nv_verror (int error_flag, const char *fmt, _BSD_VA_LIST_ arglist)
{
	int	local_errno;
	char	nv_error_string[MAX_ERROR_STRING];
	char	*p = nv_error_string;

	dprintf ("nv_verror\n");

	local_errno = error_flag & ~NV_ERROR_MASK;

	p = nv_error_string;

	if (!(error_flag & NV_ERROR_MESSAGE))
		p += sprintf (p, "%s: ",
			(error_flag & NV_ERROR_WARNING) ? "WARNING" : "ERROR");

	p += vsprintf (p, fmt, arglist);

	if ((error_flag & NV_ERROR_ERRNO) && (local_errno != 0)) {
		p += sprintf (p, " (errno=%d)\n", local_errno);
	} else if (error_flag & NV_ERROR_STATUS) {
		p += sprintf (p, " (%s)\n", nv_rm_status_text(local_errno));
	} else {
		p += sprintf (p, "\n");
	}

	out_string (nv_error_string);

	if (error_flag & NV_ERROR_PANIC) {
		nv_error (0, "fatal error");
/*		osDbgBreakPoint();*/
	}

	return (p - nv_error_string);
}

int
nv_error (int error_flag, const char *fmt, ...)
{
	va_list		arglist;
	int		chars_written;
	
	dprintf ("nv_error\n");
	va_start (arglist, fmt);
	chars_written = nv_verror (error_flag, fmt, arglist);
	va_end (arglist);

	return (chars_written);
}

void
nv_panic (const char *fmt, ...)
{
	va_list	arglist;

	dprintf ("nv_panic\n");
	va_start (arglist, fmt);
	nv_verror (NV_ERROR_PANIC | 1, fmt, arglist);
	panic ("nv_panic");
}

void
nv_warning (const char *fmt, ...)
{
	va_list	arglist;

	dprintf ("nv_warning\n");
	va_start (arglist, fmt);
	nv_verror (NV_ERROR_WARNING, fmt, arglist);
	va_end (fmt);
}

void
nv_message (const char *fmt, ...)
{
	va_list	arglist;

	dprintf ("nv_message\n");
	va_start (arglist, fmt);
	nv_verror (NV_ERROR_MESSAGE, fmt, arglist);
	va_end (arglist);
}

int
nv_printf (const char *fmt, ...)
{
	char	nv_error_string[MAX_ERROR_STRING];
	char	*p = nv_error_string;
	int	chars_written;
	va_list	arglist;

	va_start (arglist, fmt);
	chars_written = vsprintf (p, fmt, arglist);
	va_end (arglist);

	out_string (p);

	return (chars_written);
}

int
nv_debug (int debuglevel, const char *fmt, ...)
{
	char	nv_error_string[MAX_ERROR_STRING];
	char	*p = nv_error_string;
	int	c = 0;
	va_list	arglist;

	dprintf ("nv_debug\n");
	if (debuglevel >= cur_debuglevel) {
		va_start (arglist, fmt);
		c = vsprintf (p, fmt, arglist);
		va_end (arglist);
		out_string (p);
	}

	return c;
}

BOOL
osPciDevicePresent (unsigned short vendor, unsigned short device)
{
	devclass_t pci = devclass_find ("pci");
	device_t bus, dev = 0, *kids;
	int busnum, numkids, i;

	dprintf ("osPciDevicePresent: %08x %08x\n", vendor, device);
	for (busnum = 0; busnum < devclass_get_maxunit (pci); ++busnum) {
		bus = devclass_get_device (pci, busnum);
		if (!bus)
			continue;
		device_get_children (bus, &kids, &numkids);
		for (i = 0; i < numkids; ++i) {
			dev = kids[i];
			if (	pci_get_vendor (dev) == vendor &&
				pci_get_device (dev) == device) {
				free (kids, M_TEMP);
				return (TRUE);
			}
		}
		free (kids, M_TEMP);
	}
	return (FALSE);
}

void *
osPciInitHandle (U008 bus, U008 slot, U008 function, U016 *vendor, U016 *device)
{
	devclass_t pci = devclass_find ("pci");
	device_t _bus, dev = 0, *kids;
	int busnum, numkids, i;

	dprintf ("osPciInitHandle: %x %x %x\n", bus, slot, function);
	return (0);
	for (busnum = 0; busnum < devclass_get_maxunit (pci); ++busnum) {
		_bus = devclass_get_device (pci, busnum);
		if (!_bus)
			continue;
		device_get_children (_bus, &kids, &numkids);
		for (i = 0; i < numkids; ++i) {
			dev = kids[i];
			if (	(pci_get_bus (dev) == bus) &&
				(pci_get_slot (dev) == slot) &&
				(pci_get_function (dev) == function)) {
				free (kids, M_TEMP);
				*vendor = pci_get_vendor (dev);
				*device = pci_get_device (dev);
				return ((void *)dev);
			}
			
		}
		free (kids, M_TEMP);
	}
	return (0);
}

U008
osPciReadByte (void *handle, U008 offset)
{
	dprintf ("osPciReadByte");
	return ((U008) pci_read_config ((device_t)handle, offset, 1));
}

U016
osPciReadWord (void *handle, U008 offset)
{
	dprintf ("osPciReadWord\n");
	return ((U016) pci_read_config ((device_t)handle, offset, 2));
}

U032
osPciReadDword (void * handle, U008 offset)
{
	dprintf ("osPciReadDword\n");
	return ((U032) pci_read_config ((device_t)handle, offset, 4));
}

VOID
osPciWriteByte (void *handle, U008 offset, U008 value)
{
	dprintf ("osPciWriteByte\n");
	pci_write_config ((device_t)handle, offset, value, 1);
}

VOID
osPciWriteWord (void * handle, U008 offset, U016 value)
{
	dprintf ("osPciWriteWord\n");
	pci_write_config ((device_t)handle, offset, value, 2);
}

VOID
osPciWriteDword (void * handle, U008 offset, U032 value)
{
	dprintf ("osPciWriteDword\n");
	pci_write_config ((device_t)handle, offset, value, 4);
}

void
osIoWriteByte (PHWINFO pdev, unsigned long address, unsigned char value)
{
	dprintf ("osIoWriteByte\n");
	outb (address, value);
}

void
osIoWriteWord (PHWINFO pdev, unsigned long address, unsigned short value)
{
	dprintf ("osIoWriteWord\n");
	outw (address, value);
}

void
osIoWriteDword (PHWINFO pdev, unsigned long address, unsigned int value)
{
	dprintf ("owIoWriteDword\n");
	outl (address, value);
}

unsigned char
osIoReadByte (PHWINFO pdev, unsigned long address)
{
	dprintf ("osIoReadByte\n");
	return (inb (address));
}

unsigned short
osIoReadWord (PHWINFO pdev, unsigned long address)
{
	dprintf ("osIoReadWord\n");
	return (inw (address)); 
}

unsigned int
osIoReadDword (PHWINFO pdev, unsigned long address)
{
	dprintf ("osIoReadDword\n");
	return (inl (address));
}

unsigned int
osCli (unsigned int flags)
{

/*	printf ("osCli");

	flags = read_eflags();
	disable_intr ();

	return (flags);
 */
	return (0);
}

unsigned int
osSti (unsigned int flags)
{
/*
	write_eflags (flags);
	return (flags);
*/
	return (0);
}

typedef struct wb_mem_t {
	unsigned int start;
	unsigned int size;
	struct wb_mem_t *next;
} wb_mem;

static wb_mem *wb_list = NULL;

int save_wb (unsigned int start, unsigned int size);
int del_wb (unsigned int start, unsigned int size);

int
save_wb (unsigned int start, unsigned int size)
{
	wb_mem	*tmp = malloc (sizeof(wb_mem), M_NV, M_NOWAIT);

	dprintf ("save_wb\n");
	return (0);

	if (tmp == NULL)
		return (0);

	tmp->start = start;
	tmp->size = size;
	tmp->next = wb_list;

	wb_list = tmp;

	return (1);
}

int
del_wb (unsigned int start, unsigned int size)
{
	wb_mem	*tmp, *prev = NULL;

	dprintf ("del_wb\n");

	return (0);

	for (tmp = wb_list; tmp && tmp->start != start; tmp = tmp->next) {
		prev = tmp;
	}

	if (tmp == NULL)
		return (0);

	if (prev == NULL) { /* at the head */
		wb_list = tmp->next;
	} else {
		prev->next = tmp->next;
	}

	free (tmp, M_NV);
	return (1);
}

void *
osMapKernelSpace (unsigned int start, unsigned int size_bytes, U032 mode)
{
	dprintf ("osMapKernelSpace\n");
	return 0;
}

void
osUnmapKernelSpace (void *addr, unsigned int size_bytes)
{
	dprintf ("osUnmapKernelSpace\n");
	return;
}

void
osUnmapUserSpace (void *uaddr, void *priv)
{
	dprintf ("osUnmapUserSpace\n");
}

void *
osMapIOSpace (U032 start, U032 size_bytes, void **priv, U032 user, U032 mode)
{
	return NULL;
}

RM_STATUS
osMapIOPortSpace(PHWINFO pDev, U032 Address, U032 Length, U032 *uVirtualAddress
)
{
    *uVirtualAddress = Address;

    return RM_OK;
}

void
osUnmapIOSpace (void *addr, U032 size_bytes, void *priv, U032 user)
{
	return;
}

RM_STATUS
osAllocPool (void **addr, unsigned int size_bytes)
{
	dprintf ("osAllocPool\n");
	return (RM_ERROR);
}

RM_STATUS
osFreePool (void *addr)
{
	dprintf ("osFreePool\n");
	return (RM_ERROR);
}

static unsigned long contigsize = 0;

void *
osAllocContigPages (unsigned long size) 
{
	void *addr;

	dprintf ("osAllocContigPages\n");

	if (contigsize != 0)
		nv_warning ("contigsize != 0\n");

	addr = contigmalloc(size, M_NV, M_NOWAIT, 0, 0xffffffff, PAGE_SIZE, 0);
	contigsize = size;

	/* what do do in case of failure ? */
	return (addr);
}

void
osFreeContigPages (void *addr)
{
	dprintf ("osFreeContigPages\n");
	contigfree (addr, contigsize, M_NV);
}

void *
osMapUserSpace (void *kaddr, void **priv, unsigned int size_bytes, U032 mode)
{
	dprintf ("osMapUserSpace\n");
	return (NULL);	
}

U032
osGetKernPhysAddr (void *addr)
{
	dprintf ("osGetKernPhysAddr\n");
	return (U032) vtophys ((vm_offset_t)addr);
}

void * osRegistryLookup (PHWINFO, char *, char *);

void *
osRegistryLookup (PHWINFO pDev, char *regDevNode, char *regParmStr)
{
	dprintf ("osRegistryLookup\n");
	return (NULL);
}

RM_STATUS
osReadRegistryBoolean (PHWINFO pDev, char *regDevNode, char *regParmStr)
{
	return 1;
}

RM_STATUS
osReadRegistryBinary (PHWINFO pDev, char *regDevNode, char *regParmStr,
		U008 *data, U032 *cblen)
{
	dprintf ("osReadRegistryBinary\n");
	return (RM_ERROR);
}

RM_STATUS
osWriteRegistryBinary (PHWINFO pDev, char *regDevNode, char *regParmStr,
		U008 *data, U032 cblen)
{
	dprintf ("osWriteRegistryBinary\n");
	return (RM_ERROR);
}

RM_STATUS
osWriteRegistryDword (PHWINFO pDev, char *regDevNode, char *regParmStr,
	U032 cblen)
{
	dprintf ("osWriteRegistryDword\n");
	return (RM_ERROR);
}

RM_STATUS
osReadRegistryDword (PHWINFO pDev, char *regDevNode, char *regParmStr,
		U032 *cblen)
{
	dvprintf ("osReadRegistryDword\n");
	return (RM_ERROR);
}

BOOL
KernInitAGP (VOID **ap_phys_base, VOID **ap_mapped_base, U032 *apsize)
{
	return 1;
}

BOOL
KernTeardownAGP (void)
{
	return 1;
}

RM_STATUS
KernAllocAGPPages(VOID **pAddress, U032 PageCount, VOID **priv_data)
{
	*pAddress = (void*) 0;
	return RM_ERROR;
};

RM_STATUS
KernFreeAGPPages (VOID **pAddress, VOID *priv_data)
{
	return RM_OK;
}

RM_STATUS
osInitSwapBarrier (int id)
{
	return RM_OK;
}

RM_STATUS
osRemoveSwapBarrier (int id)
{
        return RM_OK;
}

RM_STATUS
osSwapBarrier (int id, VOID (*callback)(VOID *), VOID *data)
{
	return RM_OK;
}
